// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test1pushL.cpp
// Purpose:	Time Measurements for Rule Implementations: Push,local,no goto
// Last Change:	04.02.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Test Program 1:
//
// p3(X, Z) :- p2(X, Y), q(Y, Z).
// p2(X, Z) :- p1(X, Y), q(Y, Z).
// p1(X, Z) :- p0(X, Y), q(Y, Z).
// p0(X, Y) :- r(Y, Z).
//
// q = {(i, i+1) | 1 <= i <= 5000}
// r = {(i, i)   | 1 <= i <= 5000}


#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int Q_TUPLES = 5000;

const int R_TUPLES = 5000;

const bool PRINT_EXTENSION = false;

const int STACKSIZE = 50;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


// ----------------------------------------------------------------------------
// Stack (of int):
// ----------------------------------------------------------------------------


class stack {

private:
	static int stack_arr[STACKSIZE];
	static int *stack_p;
	static int stack_free;

public:
	static inline void stack_init()
	{
		stack_p = stack_arr;
		stack_free = STACKSIZE;
	}

	static inline void push(int i)
	{
		if(stack_free > 0) {
			stack_free--;
			*(stack_p++) = i;
		}
		else {
			cout << "Stack overflow\n";
			exit(1);
		}
	}

	static inline int pop()
	{
		if(stack_p > stack_arr) {
			stack_free++;
			return *(--stack_p);
		}
		else {
			cout << "Stack underflow\n";
			exit(1);
		}
	}

	static inline bool is_empty()
	{
		return (stack_p == stack_arr);
	}
};

int stack::stack_arr[STACKSIZE];
int *stack::stack_p;
int stack::stack_free;


//-----------------------------------------------------------------------------
// Predicate q: Tuples generated by computation
//-----------------------------------------------------------------------------

class q_cursor {
public:
	q_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < Q_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return i+1; }

	void close() {}

	void push() { stack::push(i); }

	void pop() { i = stack::pop(); }

private:
	int i;
};

//-----------------------------------------------------------------------------
// Predicate r: Tuples generated by computation
//-----------------------------------------------------------------------------

class r_cursor {
public:
	r_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < R_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return i; }

	void close() {}

	void push() { stack::push(i); }

	void pop() { i = stack::pop(); }

private:
	int i;
};

// ----------------------------------------------------------------------------
// Predicate p3 - Push Method
// ----------------------------------------------------------------------------


class p3_cursor {
private:

	// Continuation labels:
	typedef enum {
		TASK_init,
		TASK_CONT_r,
		TASK_START_p0,
		TASK_CONT_p0,
		TASK_START_p1,
		TASK_CONT_p1,
		TASK_START_p2,
		TASK_CONT_p2
	} task_t;

	// Functions for using the global stack:
	inline void push_int(int i)
	{
		stack::push(i);
	}

	inline int pop_int()
	{
		return stack::pop();
	}

	inline void push_task(task_t t)
	{
		stack::push(t);
	}

	inline task_t pop_task()
	{
		return (task_t) stack::pop();
	}

public:
	void open()
	{
		stack::stack_init();
		push_task(TASK_init);
	}

	bool fetch()
	{
	// Arguments:
	int p0_1, p0_2;
	int p1_1, p1_2;
	int p2_1, p2_2;

	// Cursors:
	r_cursor r_c;
	q_cursor q_c1;
	q_cursor q_c2;
	q_cursor q_c3;

	while(!stack::is_empty()) {
		switch(pop_task()) {

		case TASK_init:
			// p0(X, Y) :- r(X, Y).
			r_c.open();
			while(r_c.fetch()) {
				p0_1 = r_c.col_1();
				p0_2 = r_c.col_2();
				if(r_c.fetch()) {
					r_c.push();
					push_task(TASK_CONT_r);
				}
				// goto p0;
				push_task(TASK_START_p0);
				break;
			}
			break;

		case TASK_CONT_r:
			r_c.pop();
			do {
				p0_1 = r_c.col_1();
				p0_2 = r_c.col_2();
				if(r_c.fetch()) {
					r_c.push();
					push_task(TASK_CONT_r);
				}
				// goto p0;
				push_task(TASK_START_p0);
				break;
			} while(r_c.fetch());
			break;

		case TASK_START_p0:
			// p1(X, Z) :- p0(X, Y), q(Y, Z).
			q_c1.open();
			while(q_c1.fetch()) {
				if(q_c1.col_1() == p0_2) { // Join over Y
					p1_1 = p0_1;
					p1_2 = q_c1.col_2();
					if(q_c1.fetch()) {
						push_int(p0_1);
						push_int(p0_2);
						q_c1.push();
						push_task(TASK_CONT_p0);
					}
					// goto p1;
					push_task(TASK_START_p1);
					break;
				}
			}
			break;

		case TASK_CONT_p0:
			q_c1.pop();
			p0_2 = pop_int();
			p0_1 = pop_int();
			do {
				if(q_c1.col_1() == p0_2) { // Join over Y
					p1_1 = p0_1;
					p1_2 = q_c1.col_2();
					if(q_c1.fetch()) {
						push_int(p0_1);
						push_int(p0_2);
						q_c1.push();
						push_task(TASK_CONT_p0);
					}
					// goto p1;
					push_task(TASK_START_p1);
					break;
				}
			} while(q_c1.fetch());
			break;

		case TASK_START_p1:
			// p2(X, Z) :- p1(X, Y), q(Y, Z).
			q_c2.open();
			while(q_c2.fetch()) {
				if(q_c2.col_1() == p1_2) { // Join over Y
					p2_1 = p1_1;
					p2_2 = q_c2.col_2();
					if(q_c2.fetch()) {
						push_int(p1_1);
						push_int(p1_2);
						q_c2.push();
						push_task(TASK_CONT_p1);
					}
					// goto p2;
					push_task(TASK_START_p2);
					break;
				}
			}
			break;

		case TASK_CONT_p1:
			q_c2.pop();
			p1_2 = pop_int();
			p1_1 = pop_int();
			do {
				if(q_c2.col_1() == p1_2) { // Join over Y
					p2_1 = p1_1;
					p2_2 = q_c2.col_2();
					if(q_c2.fetch()) {
						push_int(p1_1);
						push_int(p1_2);
						q_c2.push();
						push_task(TASK_CONT_p1);
					}
					// goto p2;
					push_task(TASK_START_p2);
					break;
				}
			} while(q_c2.fetch());
			break;

		case TASK_START_p2:
			// p3(X, Z) :- p2(X, Y), q(Y, Z).
			q_c3.open();
			while(q_c3.fetch()) {
				if(q_c3.col_1() == p2_2) { // Join over Y
					p3_1 = p2_1;
					p3_2 = q_c3.col_2();
					if(q_c3.fetch()) {
						push_int(p2_1);
						push_int(p2_2);
						q_c3.push();
						push_task(TASK_CONT_p2);
					}
					// goto p3;
					return true;
				}
			}
			break;

		case TASK_CONT_p2:
			q_c3.pop();
			p2_2 = pop_int();
			p2_1 = pop_int();
			do {
				if(q_c3.col_1() == p2_2) { // Join over Y
					p3_1 = p2_1;
					p3_2 = q_c3.col_2();
					if(q_c3.fetch()) {
						push_int(p2_1);
						push_int(p2_2);
						q_c3.push();
						push_task(TASK_CONT_p2);
					}
					// goto p3;
					return true;
				}
			} while(q_c3.fetch());
			break;

		default:
			cout << "Impossible cont\n";
			return false;
		}
	}
	return false;

	}

    int col_1() { return p3_1; }
    int col_2() { return p3_2; }

    void close()
    {
    }
private:
	// Arguments:
	int p3_1, p3_2;
};


//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1 = 0;
	int c2 = 0;

	cout << "Test of Method 3: Push with local vars [Program 1]\n";
	t.start();
	p3_cursor p3_c;
	p3_c.open();
	while(p3_c.fetch()) {
		c1 += p3_c.col_1();
		c2 += p3_c.col_2();
	}
	p3_c.close();
	t.end();

	if(PRINT_EXTENSION) {
		int num_tuples = 0;
		cout << "Extension of p3:\n\n";
		p3_c.open();
		while(p3_c.fetch()) {
			num_tuples++;
			cout << "p3(" << p3_c.col_1() << ", " << p3_c.col_2()
				<< ").\n";
		}
		cout << "\n\nTotal number of tuples: " << num_tuples << "\n";
		p3_c.close();
	}

	cout << "Column sums: col1=" << c1 << ", col2=" << c2 << "\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n\n\n";

	return 0;
}

